from os import name as os_name
from shutil import which

from .common import Player


class WMPlayer(Player):
    @staticmethod
    def detect_executable() -> str:
        if os_name == "nt":
            path = r"C:\Program Files (x86)\Windows Media Player\wmplayer.exe"
            if which(path):
                return path

    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        args = [override_executable or self.executable, url]
        if full_screen:
            args.append("/fullscreen")
        return args



# Based on https://github.com/beautifier/js-beautify/blob/main/python/jsbeautifier/unpackers/packer.py
from re import compile, DOTALL, ASCII

class UnpackingError(Exception):
    pass

DETECT_PATTERN = compile(
    r"eval[ ]*\([ ]*function[ ]*\([ ]*p[ ]*,[ ]*a[ ]*,[ ]*c[ ]*,[ ]*k[ ]*,[ ]*e[ ]*,[ ]*"
)
FILTERARGS_PATTERNS = [
    compile(r"}\('(.*)', *(\d+|\[\]), *(\d+), *'(.*)'\.split\('\|'\), *(\d+), *(.*)\)\)", DOTALL),
    compile(r"}\('(.*)', *(\d+|\[\]), *(\d+), *'(.*)'\.split\('\|'\)", DOTALL),
]
WORD_PATTERN = compile(r"\b\w+\b", ASCII)
REPLACESTRINGS_PATTERN = compile(r'var *(_\w+)\=\["(.*?)"\];', DOTALL)

def detect(source: str) -> tuple[bool, str, str]:
    """Detects whether `source` is P.A.C.K.E.R. coded."""
    beginstr = ""
    endstr = ""
    begin_offset = -1
    mystr = DETECT_PATTERN.search(source)
    if mystr:
        begin_offset = mystr.start()
        beginstr = source[:begin_offset]
    if begin_offset != -1:
        source_end = source[begin_offset:]
        if source_end.split("')))", 1)[0] == source_end:
            try:
                endstr = source_end.split("}))", 1)[1]
            except IndexError:
                endstr = ""
        else:
            endstr = source_end.split("')))", 1)[1]
    return mystr is not None, beginstr, endstr

def unpack(source: str, beginstr: str = "", endstr: str = "") -> str:
    """Unpacks P.A.C.K.E.R. packed js code."""
    payload, symtab, radix, count = _filterargs(source)

    if count != len(symtab):
        raise UnpackingError("Malformed p.a.c.k.e.r. symtab.")

    try:
        unbase = Unbaser(radix)
    except TypeError:
        raise UnpackingError("Unknown p.a.c.k.e.r. encoding.")

    def lookup(match) -> str:
        """Look up symbols in the synthetic symtab."""
        word = match.group(0)
        return symtab[unbase(word)] or word

    payload = payload.replace("\\\\", "\\").replace("\\'", "'")
    source = WORD_PATTERN.sub(lookup, payload)
    return _replacestrings(source, beginstr, endstr)

def _filterargs(source: str) -> tuple[str, list[str], int, int]:
    """Juice from a source file the four args needed by decoder."""
    for juicer in FILTERARGS_PATTERNS:
        args = juicer.search(source)
        if args:
            a = args.groups()
            if a[1] == "[]":
                a = list(a)
                a[1] = 62
                a = tuple(a)
            try:
                return a[0], a[3].split("|"), int(a[1]), int(a[2])
            except ValueError:
                raise UnpackingError("Corrupted p.a.c.k.e.r. data.")
    raise UnpackingError("Could not make sense of p.a.c.k.e.r data (unexpected code structure)")

def _replacestrings(source: str, beginstr: str = "", endstr: str = "") -> str:
    """Strip string lookup table (list) and replace values in source."""
    match = REPLACESTRINGS_PATTERN.search(source)
    if match:
        varname, strings = match.groups()
        startpoint = len(match.group(0))
        lookup = strings.split('","')
        variable = "%s[%%d]" % varname
        for index, value in enumerate(lookup):
            source = source.replace(variable % index, '"%s"' % value)
        return source[startpoint:]
    return beginstr + source + endstr

class Unbaser:
    """Functor for a given base. Will efficiently convert strings to natural numbers."""
    ALPHABET = {
        62: "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ",
        95: (
            " !\"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ"
            "[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~"
        ),
    }

    def __init__(self, base: int):
        self.base = base

        if 36 < base < 62:
            if base not in self.ALPHABET:
                self.ALPHABET[base] = self.ALPHABET[62][:base]

        if 2 <= base <= 36:
            self.unbase = lambda string: int(string, base)
        else:
            try:
                self.dictionary = {cipher: index for index, cipher in enumerate(self.ALPHABET[base])}
            except KeyError:
                raise TypeError("Unsupported base encoding.")

            self.unbase = self._dictunbaser

    def __call__(self, string: str) -> int:
        return self.unbase(string)

    def _dictunbaser(self, string: str) -> int:
        """Decodes a value to an integer."""
        ret = 0
        for index, cipher in enumerate(string[::-1]):
            ret += (self.base**index) * self.dictionary[cipher]
        return ret



from .common import Player


# This is just that you can check if the player is an Android player
class AndroidPlayer(Player):
    @classmethod
    def is_available(cls) -> bool:
        return True


class AndroidChoosePlayer(AndroidPlayer):
    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        return [
            "am",
            "start",
            "--user",
            "0",
            "-a" "android.intent.action.VIEW",
            "-d",
            url,
            "-c",
            "android.intent.category.BROWSABLE",
        ]


# http://mpv-android.github.io/mpv-android/intent.html
class AndroidMPVPlayer(AndroidPlayer):
    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        return [
            "am",
            "start",
            "--user",
            "0",
            "-a" "android.intent.action.VIEW",
            "-d",
            url,
            "-n",
            "is.xyz.mpv/.MPVActivity",
        ]


# https://wiki.videolan.org/Android_Player_Intents/
class AndroidVLCPlayer(AndroidPlayer):
    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        args = [
            "am",
            "start",
            "--user",
            "0",
            "-a" "android.intent.action.VIEW",
            "-d",
            url,
            "-n",
            "org.videolan.vlc/org.videolan.vlc.gui.video.VideoPlayerActivity",
        ]
        if title:
            args.append("-e")
            args.append("title")
            args.append(title)
        return args



from re import compile as re_compile

from httpx import AsyncClient

from .common import DirectLink, Hoster

LULUVODO_PATTERN = re_compile(r'file:\s*"([^"]+)"')

class LuluvdoHoster(Hoster):
    requires_headers = True

    async def get_direct_link(self) -> DirectLink:
        async with AsyncClient(verify=False, follow_redirects=True, headers={"user-agent": "Mozilla/5.0 (Android 15; Mobile; rv:132.0) Gecko/132.0 Firefox/132.0"}) as client:
            response = await client.get(self.url)
            luluvdo_id = response.url.path.split('/')[-1]
            url = f"https://luluvdo.com/dl?op=embed&file_code={luluvdo_id}"
            response = await client.get(url)
            match = LULUVODO_PATTERN.search(response.text)
            return DirectLink(
                url=match.group(1),
                headers={"user-agent": client.headers["user-agent"]},
            )



from re import compile as re_compile

from ..networking import AsyncClient
from .common import DirectLink, Hoster

STREAMTAPE_PATTERN = re_compile(r"'botlink.*innerHTML.*?'(?P<s1>.*)'.*?\+.*?'(?P<s2>.*)'")


class StreamtapeHoster(Hoster):
    async def get_direct_link(self) -> DirectLink:
        # TODO: Error checking
        async with AsyncClient(verify=False) as client:
            response = await client.get(self.url)

            # TODO: Save html and error in order to investigate
            # with open("out.txt", "wb") as f:
            #    f.write(response.text.encode('utf-8'))

            match = STREAMTAPE_PATTERN.search(response.text)
            return DirectLink(f"https:{match.group('s1')}{match.group('s2')[4:]}")



import sys
from os import name as os_name

from .android import AndroidChoosePlayer, AndroidMPVPlayer, AndroidVLCPlayer
from .ffplay import FFPlayPlayer
from .flatpak import FlatpakCelluloidPlayer, FlatpakMPVPlayer, FlatpakVLCPlayer
from .mpv import CelluloidPlayer, MPVNETPlayer, MPVPlayer
from .vlc import VLCPlayer
from .wmplayer import WMPlayer

is_android = hasattr(sys, "getandroidapilevel")


def get_players():
    if is_android:
        return [AndroidMPVPlayer, AndroidVLCPlayer, AndroidChoosePlayer]
    _all = [MPVPlayer, VLCPlayer, FFPlayPlayer]

    if os_name == "nt":
        return _all + [MPVNETPlayer, WMPlayer]

    if os_name == "posix":
        return _all + [
            CelluloidPlayer,
            FlatpakMPVPlayer,
            FlatpakCelluloidPlayer,
            FlatpakVLCPlayer,
        ]

    return _all


available_players = get_players()

available_players_keys = []
for p in available_players:
    available_players_keys.append(p.__name__)

all_players = [
    AndroidChoosePlayer,
    AndroidMPVPlayer,
    AndroidVLCPlayer,
    FFPlayPlayer,
    MPVPlayer,
    MPVNETPlayer,
    CelluloidPlayer,
    VLCPlayer,
    WMPlayer,
    FlatpakMPVPlayer,
    FlatpakVLCPlayer,
    FlatpakCelluloidPlayer,
]

all_players_keys = []
for p in all_players:
    all_players_keys.append(p.__name__)

player_map = {}
for p in all_players:
    player_map[p.__name__] = p



from os import name as os_name
from shutil import which

from .common import Player


class VLCPlayer(Player):

    @staticmethod
    def detect_executable() -> str:
        if which("vlc"):
            return "vlc"
        if os_name == "nt":
            path_64bit = r"C:\Program Files\VideoLAN\VLC\vlc.exe"
            if which(path_64bit):
                return path_64bit
            path_32bit = r"C:\Program Files (x86)\VideoLAN\VLC\vlc.exe"
            if which(path_32bit):
                return path_32bit

    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        args = [
            override_executable or self.executable,
            url,
            "--no-video-title-show",
            "--play-and-exit",
        ]
        if full_screen:
            args.append("--fullscreen")
        if title:
            args.append(f"--input-title-format={title}")
        if headers:
            # Hehe
            if headers.get("Referer"):
                args.append(f"--http-referrer={headers.get('Referer')}")
            if headers.get("referer"):
                args.append(f"--http-referrer={headers.get('referer')}")
            if headers.get("User-Agent"):
                args.append(f"--http-user-agent={headers.get('User-Agent')}")
            if headers.get("user-agent"):
                args.append(f"--http-user-agent={headers.get('user-agent')}")
            if headers.get("User-agent"):
                args.append(f"--http-user-agent={headers.get('User-agent')}")
            if headers.get("user-Agent"):
                args.append(f"--http-user-agent={headers.get('user-Agent')}")
        return args



import logging
from asyncio import gather
from dataclasses import dataclass
from typing import Union

from bs4 import BeautifulSoup

from ..networking import AsyncClient, AcceptLanguage
from ..hoster.veo import VOEHoster
from ..hoster.vidoza import VidozaHoster
from ..hoster.speedfiles import SpeedFilesHoster
from ..hoster.doodstream import DoodstreamHoster
from ..hoster.vidmoly import VidmolyHoster
from ..hoster.filemoon import FilemoonHoster
from ..hoster.luluvdo import LuluvdoHoster
from ..hoster.streamtape import StreamtapeHoster
from ..hoster.loadx import LoadXHoster
from .common import Episode, Hoster, Language, Provider, SearchResult, Series
from ..utils import json_loads
from ..utils import fully_unescape

HOSTER_MAP = {
    "VOE": VOEHoster,
    "Doodstream": DoodstreamHoster,
    "Vidoza": VidozaHoster,
    "Streamtape": StreamtapeHoster,
    "SpeedFiles": SpeedFilesHoster,
    "Filemoon": FilemoonHoster,
    "Luluvdo": LuluvdoHoster,
    "Vidmoly": VidmolyHoster,
    "LoadX": LoadXHoster
}

LANGUAGE_MAP = {
    "english": Language.EN,
    "english-german": Language.EN_DESUB,
    "japanese-english": Language.JP_ENSUB,
    "japanese-german": Language.JP_DESUB,
    "german": Language.DE
}

def provider_to_hoster(provider: str, url: str) -> Hoster:
    if provider in HOSTER_MAP:
        return HOSTER_MAP[provider](url)


def lang_img_src_lang_name_to_lang(name: str) -> Language:
    return LANGUAGE_MAP[name]


@dataclass
class AniWorldEpisode(Episode):
    url: str

    async def process_hoster(self) -> dict[Language, list[Hoster]]:
        async with AsyncClient(accept_language=AcceptLanguage.DE) as client:
            response = await client.get(
                f"{self.url}/staffel-{self.season}/episode-{self.episode_number}"
            )
            soup = BeautifulSoup(response.text, "html.parser")
            watch_episode = soup.find_all(
                "li",
                class_=lambda value: value and value.startswith("episodeLink"),
                attrs={
                    "data-lang-key": True,
                    "data-link-id": True,
                    "data-link-target": True,
                    "data-external-embed": True,
                },
            )
            processed_hoster = {}
            for l in Language:
                processed_hoster[l] = list()

            langs = soup.find("div", class_="changeLanguage").find_all("img")
            lang_map = {}
            for lang in langs:
                lang_name = lang.attrs.get("src").split("/")[-1].rsplit(".", 1)[0]
                data_lang_key = lang.attrs.get("data-lang-key")
                lang_map[data_lang_key] = lang_img_src_lang_name_to_lang(lang_name)

            for episode in watch_episode:
                data_link_id = episode.attrs.get("data-link-id")
                provider = episode.find_next("h4").text

                data_lang_key = episode.attrs.get("data-lang-key")
                lang = lang_map[data_lang_key]

                hoster = provider_to_hoster(
                    provider, f"https://{AniWorldProvider.host}/redirect/{data_link_id}"
                )

                processed_hoster[lang].append(hoster)

            return processed_hoster


@dataclass
class AniWorldSeries(Series):
    regisseure: set[str]
    schauspieler: set[str]
    produzent: set[str]
    land: list[str]
    full_description: str
    production_year: str
    tags: set[str]

    def to_markdown(self) -> str:
        string_builder = [f"# {self.name} {self.production_year}\n{self.full_description}\n\n"]
        if len(self.regisseure) > 0:
            string_builder.append(f"- **Regisseure**: {', '.join(self.regisseure)}\n")
        if len(self.schauspieler) > 0:
            string_builder.append(f"- **Schauspieler**: {', '.join(self.schauspieler)}\n")
        if len(self.produzent) > 0:
            string_builder.append(f"- **Produzent**: {', '.join(self.produzent)}\n")
        if len(self.land) > 0:
            string_builder.append(f"- **Land**: {', '.join(self.land)}\n")
        if len(self.tags) > 0:
            string_builder.append(f"- **Tags**: {', '.join(self.tags)}")
        return "".join(string_builder)


@dataclass
class AniWorldSearchResult(SearchResult):
    link: str = None
    production_year: str = None
    host: str = None

    async def get_series(self) -> AniWorldSeries:
        return await AniWorldProvider.get_series(self)

    @property
    def url(self) -> str:
        return f"https://{self.host}/anime/stream/{self.link}"

    def __hash__(self):
        return super().__hash__()


@dataclass
class AniWorldProvider(Provider):
    host: str = "aniworld.to"

    @staticmethod
    async def search(keyword: str) -> Union[list[AniWorldSearchResult], None]:
        if keyword.strip() == "":
            return None
        async with AsyncClient(accept_language=AcceptLanguage.DE) as client:
            response = await client.get(
                f"https://{AniWorldProvider.host}/ajax/seriesSearch?keyword={keyword}"
            )
            results = json_loads(response.content)
            search_results = []
            for series in results:
                search_results.append(
                    AniWorldSearchResult(
                        provider_name="aniworld.to",
                        name=fully_unescape(series.get("name")).strip(),
                        link=series.get("link"),
                        description=fully_unescape(series.get("description")),
                        cover=f"https://{AniWorldProvider.host}{series.get('cover')}",
                        production_year=series.get("productionYear"),
                        host=AniWorldProvider.host,
                    )
                )
            return search_results

    @staticmethod
    async def get_series(search_result: AniWorldSearchResult) -> AniWorldSeries:
        async with AsyncClient(accept_language=AcceptLanguage.DE) as client:
            response = await client.get(search_result.url)
            soup = BeautifulSoup(response.text, "html.parser")

            tags = []
            for genre in soup.find_all(
                "a", class_="genreButton clearbutton", attrs={"itemprop": "genre"}
            ):
                tags.append(genre.text)

            actors = []
            for actor in soup.find_all("li", attrs={"itemprop": "actor"}):
                actors.append(actor.find_next("span", attrs={"itemprop": "name"}).text)

            creators = []
            for creator in soup.find_all("li", attrs={"itemprop": "creator"}):
                creators.append(
                    creator.find_next("span", attrs={"itemprop": "name"}).text
                )

            countrys = []
            for country in soup.find_all("li", attrs={"data-content-type": "country"}):
                countrys.append(
                    country.find_next("span", attrs={"itemprop": "name"}).text
                )

            directors = []
            for director in soup.find_all("li", attrs={"itemprop": "director"}):
                directors.append(
                    director.find_next("span", attrs={"itemprop": "name"}).text
                )

            funcs = []

            staffeln = (
                soup.find("div", attrs={"id": "stream"}, class_="hosterSiteDirectNav")
                .find("ul")
                .find_all("a")
            )
            count = 0
            for staffel in staffeln:
                # TODO: Filme
                if staffel.text != "Filme":
                    count += 1
                    if count > 1:
                        funcs.append(get_episodes_from_url(count, search_result.url))

            eps = await gather(
                get_episodes_from_soup(1, search_result.url, soup), *funcs
            )

            feps = []
            for e in eps:
                for b in e:
                    feps.append(b)

            return AniWorldSeries(
                name=fully_unescape(
                    soup.find("h1", attrs={"itemprop": "name"}).find("span").text
                ).strip(),
                production_year=fully_unescape(
                    soup.find("div", class_="series-title").find("small").text
                ).strip(),
                # age=int(soup.find("div", class_="fsk").find("span").text),
                # imdb_link=soup.find("a", class_="imdb-link").attrs.get("href"),
                full_description=fully_unescape(
                    soup.find("p", class_="seri_des").attrs.get("data-full-description")
                ),
                regisseure=directors,
                schauspieler=actors,
                produzent=creators,
                land=countrys,
                tags=tags,
                # rating_value=int(soup.find("span", attrs={"itemprop": "ratingValue"}).text),
                # rating_count=int(soup.find("span", attrs={"itemprop": "ratingCount"}).text),
                # staffeln=count,
                episodes=feps,
                cover=f"https://{search_result.host}" + soup.find("img", attrs={"itemprop": "image"}).attrs.get("data-src")
            )


async def get_episodes_from_url(staffel: int, url: str) -> list[Episode]:
    async with AsyncClient(accept_language=AcceptLanguage.DE) as client:
        response = await client.get(f"{url}/staffel-{staffel}")
        return await get_episodes_from_page(staffel, url, response.text)


async def get_episodes_from_page(staffel: int, url: str, page: str) -> list[Episode]:
    return await get_episodes_from_soup(
        staffel, url, BeautifulSoup(page, "html.parser")
    )


async def get_episodes_from_soup(
    staffel: int, url: str, soup: BeautifulSoup
) -> list[Episode]:
    episodes = []
    e_count = 0
    for episode in (
        soup.find("table", class_="seasonEpisodesList").find("tbody").find_all("tr")
    ):
        title = episode.find_next("td", class_="seasonEpisodeTitle")

        language = set()
        for flag in episode.find_next("td", class_="editFunctions").find_all("img"):
            lang_name = flag.attrs.get("src").split("/")[-1].rsplit(".", 1)[0]
            language.add(lang_img_src_lang_name_to_lang(lang_name))

        hoster = set()
        for h in episode.find_all_next("i", class_="icon"):
            t = h.attrs.get("title")
            if t in HOSTER_MAP:
                hoster.add(HOSTER_MAP[t])

        e_count += 1
        title_en = fully_unescape(title.find("span").text.strip())
        title_de = fully_unescape(title.find("strong").text.strip())
        title = (
            f"{title_en} - {title_de}"
            if title_en and title_de
            else title_en or title_de
        )
        episodes.append(
            AniWorldEpisode(
                url=url,
                title=title,
                season=staffel,
                episode_number=e_count,
                available_hoster=hoster,
                available_language=language,
                total_episode_number=None,
                # language=language,
                # hoster=hoster,
                # number=e_count,
                # staffel=staffel
            )
        )

    return episodes



from enum import Enum
from random import choice
from urllib.parse import urlparse
from typing import Union

from httpx import AsyncClient as HttpxAsyncClient, Response, AsyncBaseTransport

from rich import print
from asyncio import run


# https://www.useragents.me/
# https://github.com/yt-dlp/yt-dlp/blob/master/yt_dlp/utils/networking.py
def random_user_agent():
    _USER_AGENT_TPL = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/%s Safari/537.36'
    _CHROME_VERSIONS = (
        '90.0.4430.212',
        '90.0.4430.24',
        '90.0.4430.70',
        '90.0.4430.72',
        '90.0.4430.85',
        '90.0.4430.93',
        '91.0.4472.101',
        '91.0.4472.106',
        '91.0.4472.114',
        '91.0.4472.124',
        '91.0.4472.164',
        '91.0.4472.19',
        '91.0.4472.77',
        '92.0.4515.107',
        '92.0.4515.115',
        '92.0.4515.131',
        '92.0.4515.159',
        '92.0.4515.43',
        '93.0.4556.0',
        '93.0.4577.15',
        '93.0.4577.63',
        '93.0.4577.82',
        '94.0.4606.41',
        '94.0.4606.54',
        '94.0.4606.61',
        '94.0.4606.71',
        '94.0.4606.81',
        '94.0.4606.85',
        '95.0.4638.17',
        '95.0.4638.50',
        '95.0.4638.54',
        '95.0.4638.69',
        '95.0.4638.74',
        '96.0.4664.18',
        '96.0.4664.45',
        '96.0.4664.55',
        '96.0.4664.93',
        '97.0.4692.20',
    )
    return _USER_AGENT_TPL % choice(_CHROME_VERSIONS)

class AsyncHTTPSRedirectTransport(AsyncBaseTransport):
    async def handle_async_request(self, request) -> Response:
        url = request.url.copy_with(scheme="https")
        return Response(303, headers={"Location": str(url)})


class AcceptLanguage(Enum):
    EN = 0
    DE = 1


class AsyncClient(HttpxAsyncClient):
    def __init__(
            self,
            *args,
            http2: bool = True,
            follow_redirects: bool = True,
            auto_referer: bool = True,
            https_only: bool = True,
            accept_language: Union[AcceptLanguage, None] = AcceptLanguage.EN,
            **kwargs
    ) -> None:
        # verify=False
        self.auto_referer = auto_referer
        kwargs["http2"] = http2
        kwargs["follow_redirects"] = follow_redirects

        # aiodns / dnspython[doh]
        # socksio - SOCKS proxy support. (Optional, with httpx[socks])

        user_agent = random_user_agent()
        headers = {
            # Add others
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            # "br" = "brotli" -> httpx[brotli]
            # "zstd" -> httpx[zstd] wait for next release https://github.com/encode/httpx/pull/3139
            # "Accept-Encoding": "gzip, deflate, br", httpx is covering this
            # "Accept-Language": "en-us,en;q=0.5",  see below
            # "Host": "xxx", httpx is covering this
            # "Sec-Ch-Ua-Platform": "macOS",  # only on mac
            "Sec-Fetch-Dest": "document",
            "Sec-Fetch-Mode": "navigate",
            "Sec-Fetch-Site": "none",
            "Sec-Fetch-User": "?1",  # Not on iphone
            "Upgrade-Insecure-Requests": "1",  # Not on iphone
            "User-Agent": user_agent
            # "X-Amzn-Trace-Id": "Root=1-xxxxxxxx-xxxxxxxxxxxxxxxxxxxxxxxx"
            # TODO: More variation
        }

        if accept_language is AcceptLanguage.EN:
            headers["Accept-Language"] = "en-us,en;q=0.5"  # "en-US,en;q=0.9", "en-US"
        elif accept_language is AcceptLanguage.DE:
            headers["Accept-Language"] = choice([
                "de-DE,de;q=0.9",
                "de",  # found on macos
                "de-DE,de;q=0.9",  # found on ios
                "de-DE,de",
                "de,en-US;q=0.7,en;q=0.3"
            ])

        if kwargs.get("headers") is not None:
            headers = {**kwargs.get("headers"), **headers}
        kwargs["headers"] = headers

        if https_only is True:
            kwargs["mounts"] = {"http://": AsyncHTTPSRedirectTransport()}

        super().__init__(*args, **kwargs)

    async def request(self, *args, **kwargs) -> Response:
        if self.auto_referer is True:
            parsed_url = urlparse(args[1])  # maby use httpx.URL instead ?
            base_url = f'{parsed_url.scheme}://{parsed_url.netloc}'
            headers = {"Referer": base_url}
            if kwargs.get("headers") is not None:
                headers = {**kwargs.get("headers"), **headers}
            kwargs["headers"] = headers
        return await super().request(*args, **kwargs)


async def main():
    from .utils import json_loads
    async with AsyncClient() as client:
        response = await client.get("https://httpbin.org/headers")
        print(json_loads(response.content))
    async with HttpxAsyncClient() as client:
        response = await client.get("https://httpbin.org/headers")
        print(json_loads(response.content))

if __name__ == "__main__":
    run(main())



from dataclasses import dataclass
from random import choices
from re import compile as re_compile
from string import ascii_letters, digits
from time import time

from ..networking import AsyncClient
from .common import DirectLink, Hoster

EXTRACT_DOODSTREAM_HLS_PATTERN = re_compile(r"/pass_md5/[\w-]+/(?P<token>[\w-]+)")


def random_str(length: int = 10) -> str:
    return "".join(choices(ascii_letters + digits, k=length))


def js_date_now() -> int:
    return int(time() * 1000)


@dataclass
class DoodstreamHoster(Hoster):
    requires_headers = True

    async def get_direct_link(self) -> DirectLink:
        async with AsyncClient(verify=False, auto_referer=True) as client:
            response1 = await client.get(self.url)
            match = EXTRACT_DOODSTREAM_HLS_PATTERN.search(response1.text)

            # Require Referer
            response2 = await client.get(str(response1.url.copy_with(path=match.group())))
            return DirectLink(
                url=f"{response2.text}{random_str()}?token={match.group('token')}&expiry={js_date_now()}",
                headers={"Referer": str(response2.url.copy_with(path="/"))},
            )



from dataclasses import dataclass
from re import compile as re_compile

from ..networking import AsyncClient

from .common import DirectLink, Hoster

VIDMOLY_PATTERN = re_compile(r"sources: \[{file:\"(?P<url>.*?)\"}]")


@dataclass
class VidmolyHoster(Hoster):
    requires_headers = True

    async def get_direct_link(self) -> DirectLink:
        async with AsyncClient(verify=False, auto_referer=False) as client:
            response = await client.get(self.url)
            match = VIDMOLY_PATTERN.search(response.text)
            return DirectLink(
                url=match.group("url"),
                headers={"Referer": "https://vidmoly.to/"}
            )



from os import name as os_name
from subprocess import DEVNULL, Popen

from .common import Player
from .mpv import CelluloidPlayer, MPVPlayer
from .vlc import VLCPlayer


# This is just that you can check if the player is a Flatpak player
class FlatpakPlayer(Player):
    pass


# TODO: Dont use Popen, it will slow down


class FlatpakMPVPlayer(MPVPlayer, FlatpakPlayer):

    @classmethod
    def is_available(cls) -> bool:
        if os_name == "posix":
            p = Popen(
                ["flatpak", "info", "io.mpv.Mpv"],
                stdout=DEVNULL,
                stderr=DEVNULL,
                stdin=DEVNULL,
            )
            if p.wait() == 0:
                return True

    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        uf_args = super().play(url, title, full_screen, headers)
        uf_args.pop(0)
        return ["flatpak", "run", "io.mpv.Mpv"] + uf_args


class FlatpakCelluloidPlayer(CelluloidPlayer, FlatpakPlayer):

    @classmethod
    def is_available(cls) -> bool:
        if os_name == "posix":
            p = Popen(
                ["flatpak", "info", "io.github.celluloid_player.Celluloid"],
                stdout=DEVNULL,
                stderr=DEVNULL,
                stdin=DEVNULL,
            )
            if p.wait() == 0:
                return True

    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        uf_args = super().play(url, title, full_screen, headers)
        uf_args.pop(0)
        return ["flatpak", "run", "io.github.celluloid_player.Celluloid"] + uf_args


class FlatpakVLCPlayer(VLCPlayer, FlatpakPlayer):

    @classmethod
    def is_available(cls) -> bool:
        if os_name == "posix":
            p = Popen(
                ["flatpak", "info", "org.videolan.VLC"],
                stdout=DEVNULL,
                stderr=DEVNULL,
                stdin=DEVNULL,
            )
            if p.wait() == 0:
                return True

    def play(
        self,
        url: str,
        title: str,
        full_screen: bool,
        headers: dict[str, str] = None,
        override_executable: str = None,
    ) -> list[str]:
        uf_args = super().play(url, title, full_screen, headers)
        uf_args.pop(0)
        return ["flatpak", "run", "org.videolan.VLC"] + uf_args
